@@ -1,5 +1,6 @@
 from dataclasses import dataclass, field
 from typing import Dict, List, Set, Tuple, Union, ByteString
+from itertools import chain
 import re
 from functools import reduce
 from collections import namedtuple
@@ -64,6 +65,10 @@ class Tile:
     for this specific tile type"""
     width: int
     height: int
+    # At the time of packing/unpacking the information about the types of cells
+    # is already lost, it is critical to work through the 'logicinfo' table so
+    # store it.
+    ttyp: int
     # a mapping from dest, source wire to bit coordinates
     pips: Dict[str, Dict[str, Set[Coord]]] = field(default_factory=dict)
     clock_pips: Dict[str, Dict[str, Set[Coord]]] = field(default_factory=dict)
@@ -78,11 +83,21 @@ class Device:
     grid: List[List[Tile]] = field(default_factory=list)
     timing: Dict[str, Dict[str, List[float]]] = field(default_factory=dict)
     packages: Dict[str, Tuple[str, str, str]] = field(default_factory=dict)
-    pinout: Dict[str, Dict[str, Dict[str, str]]] = field(default_factory=dict)
+    # {variant: {package: {pin#: (pin_name, [cfgs])}}}
+    pinout: Dict[str, Dict[str, Dict[str, Tuple[str, List[str]]]]] = field(default_factory=dict)
     pin_bank: Dict[str, int] = field(default_factory = dict)
     cmd_hdr: List[ByteString] = field(default_factory=list)
     cmd_ftr: List[ByteString] = field(default_factory=list)
     template: np.ndarray = None
+    # allowable values of bel attributes
+    # {table_name: [(attr_id, attr_value)]}
+    logicinfo: Dict[str, List[Tuple[int, int]]] = field(default_factory=dict)
+    # fuses for a pair of the "features" (or pairs of parameter values)
+    # {ttype: {table_name: {(feature_A, feature_B): {bits}}}
+    shortval: Dict[int, Dict[str, Dict[Tuple[int, int], Set[Coord]]]] = field(default_factory=dict)
+    # fuses for 16 of the "features"
+    # {ttype: {table_name: {(feature_0, feature_1, ..., feature_15): {bits}}}
+    longval: Dict[int, Dict[str, Dict[Tuple[int, int, int, int, int, int, int, int, int, int, int, int, int, int, int, int], Set[Coord]]]] = field(default_factory=dict)
     # always-connected dest, src aliases
     aliases: Dict[Tuple[int, int, str], Tuple[int, int, str]] = field(default_factory=dict)
 
@@ -145,6 +160,17 @@ def fse_pips(fse, ttyp, table=2, wn=wirenames):
 
     return pips
 
+# make PLL bels
+def fse_pll(device, fse, ttyp):
+    bels = {}
+    # XXX use second grid in order to find PLL ttypes
+    if device in ['GW1N-1',  'GW1NZ-1']:
+        if ttyp == 89:
+            bel = bels.setdefault('RPLLB', Bel())
+        else:
+            bel = bels.setdefault('RPLLA', Bel())
+    return bels
+
 # add the ALU mode
 # new_mode_bits: string like "0110000010011010"
 def add_alu_mode(base_mode, modes, lut, new_alu_mode, new_mode_bits):
@@ -238,19 +264,24 @@ def fse_luts(fse, ttyp):
     # main fuse: enable shadow SRAM in the slice
     # shortval(28) [2, 0, fuses]
     if 28 in fse[ttyp]['shortval']:
-        data = fse[ttyp]['shortval'][28]
+        for i in range(6):
+            bel = luts.setdefault(f"DFF{i}", Bel())
+            mode = bel.modes.setdefault("RAM", set())
+            for key0, key1, *fuses in fse[ttyp]['shortval'][25+i//2]:
+                if key0 < 0:
+                    for f in fuses:
+                        if f == -1: break
+                        coord = fuse.fuse_lookup(fse, ttyp, f)
+                        mode.add(coord)
+
         bel = luts.setdefault(f"RAM16", Bel())
         mode = bel.modes.setdefault("0", set())
-        for key0, key1, *fuses in data:
+        for key0, key1, *fuses in fse[ttyp]['shortval'][28]:
             if key0 == 2 and key1 == 0:
-                for f in (f for f in fuses if f != -1):
+                for f in fuses:
+                    if f == -1: break
                     coord = fuse.fuse_lookup(fse, ttyp, f)
-                    mode.update({coord})
-                break
-        bel.flags.update({k:v for (k, v) in luts["LUT0"].flags.items()})
-        bel.flags.update({k+16:v for (k, v) in luts["LUT1"].flags.items()})
-        bel.flags.update({k+32:v for (k, v) in luts["LUT2"].flags.items()})
-        bel.flags.update({k+48:v for (k, v) in luts["LUT3"].flags.items()})
+                    mode.add(coord)
         bel.portmap = {
             'DI': ("A5", "B5", "C5", "D5"),
             'CLK': "CLK2",
@@ -303,37 +334,108 @@ def fse_osc(device, fse, ttyp):
 
     return osc
 
-# XXX This is not a nice way, I will replace it with automatic creation of banks
-# in the very near future.
-def set_banks(device, db):
+def set_banks(fse, db):
     # fill the bank# : corner tile table
     w = db.cols - 1
     h = db.rows - 1
-    if device == 'GW1NZ-1':
-        db.grid[0][0].bels.setdefault('BANK0', Bel())
-        db.grid[0][w].bels.setdefault('BANK1', Bel())
-    elif device == 'GW1NS-2':
-        db.grid[0][0].bels.setdefault('BANK0', Bel())
-        db.grid[0][w].bels.setdefault('BANK1', Bel())
-        db.grid[h][w].bels.setdefault('BANK2', Bel())
-        db.grid[0][0].bels.setdefault('BANK3', Bel())
-    elif device == 'GW1NS-4':
-        db.grid[0][0].bels.setdefault('BANK0', Bel())
-        db.grid[0][w].bels.setdefault('BANK1', Bel())
-        db.grid[h][w].bels.setdefault('BANK2', Bel())
-        db.grid[h][w].bels.setdefault('BANK3', Bel())
-    elif device == 'GW1N-9':
-        db.grid[0][0].bels.setdefault('BANK0', Bel())
-        db.grid[0][w].bels.setdefault('BANK1', Bel())
-        db.grid[h][w].bels.setdefault('BANK2', Bel())
-        db.grid[h][0].bels.setdefault('BANK3', Bel())
-        db.grid[0][0].bels.setdefault('BANK10', Bel())
-        db.grid[0][0].bels.setdefault('BANK30', Bel())
-    else:
-        db.grid[0][0].bels.setdefault('BANK0', Bel())
-        db.grid[0][w].bels.setdefault('BANK1', Bel())
-        db.grid[h][w].bels.setdefault('BANK2', Bel())
-        db.grid[h][0].bels.setdefault('BANK3', Bel())
+    for row, col in [(0, 0), (0, w), (h, 0), (h, w)]:
+        ttyp = fse['header']['grid'][61][row][col]
+        if 'longval' in fse[ttyp].keys():
+            if 37 in fse[ttyp]['longval'].keys():
+                for rd in fse[ttyp]['longval'][37]:
+                    db.grid[row][col].bels.setdefault(f"BANK{rd[0]}", Bel())
+
+_known_logic_tables = {
+            8:  'DCS',
+            9:  'GSR',
+            10: 'IOLOGIC',
+            11: 'IOB',
+            12: 'SLICE',
+            13: 'BSRAM',
+            14: 'DSP',
+            15: 'PLL',
+            62: 'USB',
+        }
+
+_known_tables = {
+             4: 'CONST',
+             5: 'LUT',
+            21: 'IOLOGICA',
+            22: 'IOLOGICB',
+            23: 'IOBA',
+            24: 'IOBB',
+            25: 'CLS0',
+            26: 'CLS1',
+            27: 'CLS2',
+            28: 'CLS3',
+            35: 'PLL',
+            37: 'BANK',
+            40: 'IOBC',
+            41: 'IOBD',
+            42: 'IOBE',
+            43: 'IOBF',
+            44: 'IOBG',
+            45: 'IOBH',
+            46: 'IOBI',
+            47: 'IOBJ',
+            53: 'DLLDEL0',
+            54: 'DLLDEL1',
+            56: 'DLL0',
+            64: 'USB',
+            66: 'EFLASH',
+            68: 'ADC',
+            80: 'DLL1',
+        }
+
+def fse_fill_logic_tables(dev, fse):
+    # logicinfo
+    for ltable in fse['header']['logicinfo'].keys():
+        if ltable in _known_logic_tables.keys():
+            table = dev.logicinfo.setdefault(_known_logic_tables[ltable], [])
+        else:
+            table = dev.logicinfo.setdefault(f"unknown_{ltable}", [])
+        for attr, val, _ in fse['header']['logicinfo'][ltable]:
+            table.append((attr, val))
+    # shortval
+    ttypes = {t for row in fse['header']['grid'][61] for t in row}
+    for ttyp in ttypes:
+        if 'shortval' in fse[ttyp].keys():
+            ttyp_rec = dev.shortval.setdefault(ttyp, {})
+            for stable in fse[ttyp]['shortval'].keys():
+                if stable in _known_tables:
+                    table = ttyp_rec.setdefault(_known_tables[stable], {})
+                else:
+                    table = ttyp_rec.setdefault(f"unknown_{stable}", {})
+                for f_a, f_b, *fuses in fse[ttyp]['shortval'][stable]:
+                    table[(f_a, f_b)] = {fuse.fuse_lookup(fse, ttyp, f) for f in unpad(fuses)}
+        if 'longval' in fse[ttyp].keys():
+            ttyp_rec = dev.longval.setdefault(ttyp, {})
+            for ltable in fse[ttyp]['longval'].keys():
+                if ltable in _known_tables:
+                    table = ttyp_rec.setdefault(_known_tables[ltable], {})
+                else:
+                    table = ttyp_rec.setdefault(f"unknown_{ltable}", {})
+                for f0, f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12, f13, f14, f15, *fuses in fse[ttyp]['longval'][ltable]:
+                    table[(f0, f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12, f13, f14, f15)] = {fuse.fuse_lookup(fse, ttyp, f) for f in unpad(fuses)}
+
+_pll_loc = {
+ 'GW1N-1':
+   {'TRPLL0CLK0': (0, 17, 'F4'), 'TRPLL0CLK1': (0, 17, 'F5'),
+    'TRPLL0CLK2': (0, 17, 'F6'), 'TRPLL0CLK3': (0, 17, 'F7'), },
+ 'GW1NZ-1':
+   {'TRPLL0CLK0': (0, 17, 'F4'), 'TRPLL0CLK1': (0, 17, 'F5'),
+    'TRPLL0CLK2': (0, 17, 'F6'), 'TRPLL0CLK3': (0, 17, 'F7'), },
+}
+def fse_create_pll_clock_aliases(db, device):
+    # we know exactly where the PLL is and therefore know which aliases to create
+    for row in range(db.rows):
+        for col in range(db.cols):
+            for w_dst, w_srcs in db.grid[row][col].clock_pips.items():
+                for w_src in w_srcs.keys():
+                    # XXX
+                    if device in {'GW1N-1', 'GW1NZ-1'}:
+                        if w_src in _pll_loc[device].keys():
+                            db.aliases[(row, col, w_src)] = _pll_loc[device][w_src]
 
 def from_fse(device, fse):
     dev = Device()
@@ -342,18 +444,67 @@ def from_fse(device, fse):
     for ttyp in ttypes:
         w = fse[ttyp]['width']
         h = fse[ttyp]['height']
-        tile = Tile(w, h)
+        tile = Tile(w, h, ttyp)
         tile.pips = fse_pips(fse, ttyp, 2, wirenames)
         tile.clock_pips = fse_pips(fse, ttyp, 38, clknames)
         if 5 in fse[ttyp]['shortval']:
             tile.bels = fse_luts(fse, ttyp)
         if 51 in fse[ttyp]['shortval']:
             tile.bels = fse_osc(device, fse, ttyp)
+        # XXX GW1N(Z)-1 for now
+        if ttyp in [88, 89]:
+            tile.bels = fse_pll(device, fse, ttyp)
         tiles[ttyp] = tile
 
+    fse_fill_logic_tables(dev, fse)
     dev.grid = [[tiles[ttyp] for ttyp in row] for row in fse['header']['grid'][61]]
+    fse_create_pll_clock_aliases(dev, device)
     return dev
 
+# get fuses for attr/val set using short/longval table
+# returns a bit set
+def get_table_fuses(attrs, table):
+    bits = set()
+    for key, fuses in table.items():
+        # all 16 "features" must be present to be able to use a set of bits from the record
+        have_full_key = True
+        for attrval in key:
+            if attrval == 0: # no "feature"
+                continue
+            if attrval > 0:
+                # this "feature" must present
+                if attrval not in attrs:
+                    have_full_key = False
+                    break
+                continue
+            if attrval < 0:
+                # this "feature" is set by default and can only be unset
+                if abs(attrval) in attrs:
+                    have_full_key = False
+                    break
+        if not have_full_key:
+            continue
+        bits.update(fuses)
+    return bits
+
+# get fuses for attr/val set using shortval table for ttyp
+# returns a bit set
+def get_shortval_fuses(dev, ttyp, attrs, table_name):
+    return get_table_fuses(attrs, dev.shortval[ttyp][table_name])
+
+# get fuses for attr/val set using longval table for ttyp
+# returns a bit set
+def get_longval_fuses(dev, ttyp, attrs, table_name):
+    return get_table_fuses(attrs, dev.longval[ttyp][table_name])
+
+# add the attribute/value pair into an set, which is then passed to
+# get_longval_fuses() and get_shortval_fuses()
+def add_attr_val(dev, logic_table, attrs, attr, val):
+    for idx, attr_val in enumerate(dev.logicinfo[logic_table]):
+        if attr_val[0] == attr and attr_val[1] == val:
+            attrs.add(idx)
+            break
+
 def get_pins(device):
     if device not in {"GW1N-1", "GW1NZ-1", "GW1N-4", "GW1N-9", "GW1NR-9", "GW1N-9C", "GW1NR-9C", "GW1NS-2", "GW1NS-2C", "GW1NS-4", "GW1NSR-4C"}:
         raise Exception(f"unsupported device {device}")
@@ -441,6 +592,16 @@ def json_pinout(device):
         raise Exception("unsupported device")
 
 
+
+_pll_inputs = [(5, 'CLKFB'), (6, 'FBDSEL0'), (7, 'FBDSEL1'), (8, 'FBDSEL2'), (9, 'FBDSEL3'),
+               (10, 'FBDSEL4'), (11, 'FBDSEL5'),
+               (12, 'IDSEL0'), (13, 'IDSEL1'), (14, 'IDSEL2'), (15, 'IDSEL3'), (16, 'IDSEL4'),
+               (17, 'IDSEL5'),
+               (18, 'ODSEL0'), (19, 'ODSEL1'), (20, 'ODSEL2'), (21, 'ODSEL3'), (22, 'ODSEL4'),
+               (24, 'PSDA0'), (25, 'PSDA1'), (26, 'PSDA2'), (27, 'PSDA3'),
+               (28, 'DUTYDA0'), (29, 'DUTYDA1'), (30, 'DUTYDA2'), (31, 'DUTYDA3'),
+               (32, 'FDLY0'), (33, 'FDLY1'), (34, 'FDLY2'), (35, 'FDLY3')]
+_pll_outputs = [(0, 'CLKOUT'), (1, 'LOCK'), (2, 'CLKOUTP'), (3, 'CLKOUTD'), (4, 'CLKOUTD3')]
 def dat_portmap(dat, dev):
     for row in dev.grid:
         for tile in row:
@@ -463,6 +624,28 @@ def dat_portmap(dat, dev):
                         bel.portmap['I'] = out
                         oe = wirenames[dat[f'Iobuf{pin}OE']]
                         bel.portmap['OE'] = oe
+                elif name.startswith("ODDR"):
+                        d0 = wirenames[dat[f'Iologic{pin}In'][1]]
+                        bel.portmap['D0'] = d0
+                        d1 = wirenames[dat[f'Iologic{pin}In'][2]]
+                        bel.portmap['D1'] = d1
+                        tx = wirenames[dat[f'Iologic{pin}In'][27]]
+                        bel.portmap['TX'] = tx
+                elif name == 'RPLLA':
+                    for idx, nam in _pll_inputs:
+                        wire = wirenames[dat['PllIn'][idx]]
+                        bel.portmap[nam] = wire
+                    for idx, nam in _pll_outputs:
+                        wire = wirenames[dat['PllOut'][idx]]
+                        bel.portmap[nam] = wire
+                    bel.portmap['CLKIN'] = wirenames[124];
+                elif name == 'RPLLB':
+                    reset = wirenames[dat['PllIn'][0]]
+                    bel.portmap['RESET'] = reset
+                    reset_p = wirenames[dat['PllIn'][1]]
+                    bel.portmap['RESET_P'] = reset_p
+                    odsel5 = wirenames[dat['PllIn'][23]]
+                    bel.portmap['ODSEL5'] = odsel5
 
 def dat_aliases(dat, dev):
     for row in dev.grid:
@@ -671,3 +854,4 @@ def loc2bank(db, row, col):
             bank = db.pin_bank[name + 'B']
     return bank
 
+
