@@ -1,17 +1,14 @@
 from multiprocessing.dummy import Pool
-from PIL import Image
-import numpy as np
 import pickle
 import json
 import re
 from apycula import tiled_fuzzer
 from apycula import codegen
 from apycula import pindef
-from apycula import bslib
 from apycula import chipdb
 from apycula import fuse_h4x
 from apycula import gowin_unpack
-from apycula.wirenames import wirenames, clknames, wirenumbers, clknumbers
+from apycula.wirenames import clknumbers
 
 def dff(mod, cst, row, col, clk=None):
     "make a dff with optional clock"
@@ -267,11 +264,10 @@ def spine_aliases(quads, dests, clks):
     return aliases
 
 # add border cells
-def add_rim(rows, cols):
+def add_rim(rows, cols, spine_row):
     if 1 in rows:
         rows.add(0)
-    else:
-        # XXX fill all rows
+    if max(rows) > spine_row:
         rows.update({row for row in range(max(rows) + 1, db.rows)})
     if 1 in cols:
         cols.add(0)
@@ -282,7 +278,7 @@ def add_rim(rows, cols):
 def tap_aliases(quads):
     aliases = {}
     for _, (rows, cols, spine_row) in quads.items():
-        add_rim(rows, cols)
+        add_rim(rows, cols, spine_row)
         for col in cols:
             for row in rows:
                 for src in ["GT00", "GT10"]:
@@ -293,14 +289,14 @@ def tap_aliases(quads):
 
 def branch_aliases(quads, clks):
     aliases = {}
-    for ct, (rows, _, _) in quads.items():
+    for ct, (rows, _, spine_row) in quads.items():
         for clk, taps in clks[ct].items():
             if clk < 4:
                 src = "GBO0"
             else:
                 src = "GBO1"
             for tap, branch_cols in taps.items():
-                add_rim(rows, branch_cols)
+                add_rim(rows, branch_cols, spine_row)
                 for row in rows:
                     for col in branch_cols:
                         aliases[row, col, f"GB{clk}0"] = row, tap, src
@@ -398,7 +394,7 @@ def make_lw_aliases(fse, dat, db, quads, clks):
         for lw, tap_desc in taps.items():
             for tap_col, cols in tap_desc.items():
                 tap_row = 0
-                if row > (center_row * 2):
+                if row > (center_row * 2) and has_bottom_quadrants:
                     tap_row = last_row
                 db.aliases.update({(row, tap_col, 'LT01') : (tap_row, tap_col, 'LT02')})
                 db.aliases.update({(row, tap_col, 'LT04') : (tap_row, tap_col, 'LT13')})
@@ -410,6 +406,10 @@ def make_lw_aliases(fse, dat, db, quads, clks):
     rows = { (0, 'T') }
     if has_bottom_quadrants:
         rows.update({ (last_row, 'B') })
+    else:
+        for tp in ['LT02', 'LT13']:
+            del db.grid[last_row][tap_col].pips[tp]
+
     for row, qd in rows:
         for lw, tap_desc in taps.items():
             for tap_col, cols in tap_desc.items():
@@ -424,16 +424,18 @@ def make_lw_aliases(fse, dat, db, quads, clks):
                 for tp in ['LT02', 'LT13']:
                     for pip in [p for p in db.grid[row][tap_col].pips[tp] if p not in pip2keep]:
                         del db.grid[row][tap_col].pips[tp][pip]
+
     # logic entries
     srcs = {}
-    for i, src in enumerate(dat['UfbIns']):
+    for i, src in dat['CmuxIns'].items():
         row, col, pip = src
         if pip == 126: # CLK2
-            db.aliases.update({ (center_row, col82, f'UNK{i + 104}'): (row - 1, col -1, 'CLK2')})
-            db.aliases.update({ (center_row, col81, f'UNK{i + 104}'): (row - 1, col -1, 'CLK2')})
+            idx = int(i)
+            db.aliases.update({ (center_row, col82, f'UNK{idx + 80}'): (row - 1, col -1, 'CLK2')})
+            db.aliases.update({ (center_row, col81, f'UNK{idx + 80}'): (row - 1, col -1, 'CLK2')})
             if has_bottom_quadrants:
-                db.aliases.update({ (center_row, col83, f'UNK{i + 104}'): (row - 1, col -1, 'CLK2')})
-                db.aliases.update({ (center_row, col84, f'UNK{i + 104}'): (row - 1, col -1, 'CLK2')})
+                db.aliases.update({ (center_row, col83, f'UNK{idx + 80}'): (row - 1, col -1, 'CLK2')})
+                db.aliases.update({ (center_row, col84, f'UNK{idx + 80}'): (row - 1, col -1, 'CLK2')})
 
 if __name__ == "__main__":
     quads = quadrants()
