@@ -370,8 +370,18 @@ pos = print_arr16(['PllOut'], pos, 0x5)
 pos = print_arr16(['PllInDlt'], pos, 0x24)
 pos = print_arr16(['PllOutDlt'], pos, 0x5)
 pos = print_clkins(['PllClkin'], pos, 6)
+pos = print_arr16(['SpecPll0Ins'], pos, 0x24)
+pos = print_arr16(['SpecPll0Outs'], pos, 5)
+pos = print_arr16(['SpecPll0Clkin'], pos, 6)
+pos = print_arr16(['SpecPll1Ins'], pos, 0x24)
+pos = print_arr16(['SpecPll1Outs'], pos, 5)
+pos = print_arr16(['SpecPll1Clkin'], pos, 6)
 pos = print_arr16(['DllIn'], pos, 4)
 pos = print_arr16(['DllOut'], pos, 9)
+pos = print_arr16(['SpecDll0Ins'], pos, 4)
+pos = print_arr16(['SpecDll0Outs'], pos, 9)
+pos = print_arr16(['SpecDll1Ins'], pos, 4)
+pos = print_arr16(['SpecDll1Outs'], pos, 9)
 pos = print_mult(['MultIn'], pos, 0x4f)
 pos = print_mult(['MultOut'], pos, 0x48)
 pos = print_mult(['MultInDlt'], pos, 0x4f)
@@ -390,8 +400,8 @@ pos = print_mult(['CtrlIn'], pos, 0xe)
 pos = print_mult(['CtrlInDlt'], pos, 0xe)
 print()
 #print(hex(pos))
-print(d[pos:320314 + 42628].hex())
-pos = 320314 + 42628
+print(d[pos:363662].hex())
+pos = 363662
 #print(hex(pos))
 for i in range(320):
     pos = print_arr16(['CiuConnection', i], pos, 60)
@@ -418,16 +428,13 @@ pos = print_outs(['UfbOuts'], pos, 0x20) # 20
 pos = pos + 4
 pos = print_outs(['McuIns'], pos, 0x109)
 pos = print_outs(['McuOuts'], pos, 0x174)
+pos = print_outs(['EMcuIns'], pos, 0x10e)
+pos = print_outs(['EMcuOuts'], pos, 0x13f)
 pos = print_outs(['AdcIns'], pos, 0xf)
 pos = print_outs(['AdcOuts'], pos, 13)
 pos = print_outs(['Usb2PhyIns'], pos, 0x46)
-pos = print_outs(['Usb2PhyOuts'], pos, 0x20)
+pos = print_outs(['Usb2PhyOuts'], pos, 0x2a)
 pos = print_outs(['Eflash128kIns'], pos, 0x39)
-#assert pos == 0x6f89e
-
-print(d[0x6f89e + 42628:0x6f8da + 42628].hex())
-
-pos = 0x6f8da + 42628
 pos = print_outs(['Eflash128kOuts'], pos, 0x21)
 pos = print_outs(['SpmiIns'], pos, 0x17)
 pos = print_outs(['SpmiOuts'], pos, 0x2f)
