@@ -3,6 +3,7 @@ import os
 import re
 import pickle
 import itertools
+import math
 import numpy as np
 import json
 import argparse
@@ -10,9 +11,15 @@ import importlib.resources
 from collections import namedtuple
 from apycula import codegen
 from apycula import chipdb
+from apycula.chipdb import add_attr_val, get_shortval_fuses, get_longval_fuses
+from apycula import attrids
+from apycula.attrids import pll_attrids, pll_attrvals
 from apycula import bslib
+from apycula import attrids
 from apycula.wirenames import wirenames, wirenumbers
 
+device = ""
+
 _verilog_name = re.compile(r"^[A-Za-z_0-9][A-Za-z_0-9$]*$")
 def sanitize_name(name):
     retname = name
@@ -28,8 +35,10 @@ def sanitize_name(name):
 
 def get_bels(data):
     later = []
-    belre = re.compile(r"R(\d+)C(\d+)_(?:GSR|SLICE|IOB|MUX2_LUT5|MUX2_LUT6|MUX2_LUT7|MUX2_LUT8|ODDR|OSC[ZFH]?|BUFS)(\w*)")
+    belre = re.compile(r"R(\d+)C(\d+)_(?:GSR|SLICE|IOB|MUX2_LUT5|MUX2_LUT6|MUX2_LUT7|MUX2_LUT8|ODDR|OSC[ZFH]?|BUFS|RAMW|RPLL[AB])(\w*)")
     for cellname, cell in data['modules']['top']['cells'].items():
+        if cell['type'].startswith('DUMMY_') :
+            continue
         bel = cell['attributes']['NEXTPNR_BEL']
         if bel in {"VCC", "GND"}: continue
         bels = belre.match(bel)
@@ -66,6 +75,235 @@ def get_pips(data):
 def infovaluemap(infovalue, start=2):
     return {tuple(iv[:start]):iv[start:] for iv in infovalue}
 
+# Permitted frequencies for chips
+# { device : (max_in, max_out, min_out, max_vco, min_vco) }
+_permitted_freqs = {
+        "GW1N-1":  (400, 450, 3.125,  900,  400),
+        "GW1NZ-1": (400, 400, 3.125,  800,  400),
+        "GW1N-4":  (400, 500, 3.125,  1000, 400),
+        "GW1NS-4": (400, 600, 4.6875, 1200, 600),
+        "GW1N-9":  (400, 500, 3.125,  1000, 400),
+        "GW1N-9C": (400, 600, 3.125,  1200, 400),
+        "GW1NS-2": (400, 500, 3.125,  1200, 400),
+        }
+# input params are calculated as described in GOWIN doc (UG286-1.7E_Gowin Clock User Guide)
+# fref = fclkin / idiv
+# fvco = (odiv * fdiv * fclkin) / idiv
+#
+# returns (fclkin_idx, icp, r_idx)
+# fclkin_idx - input frequency range index
+# icp - charge current
+# r_idx - resistor value index
+
+# There are not many resistors so the whole frequency range is divided into
+# 30MHz intervals and the number of this interval is one of the fuse sets. But
+# the resistor itself is not directly dependent on the input frequency.
+_freq_R = [(2.6, 65100.0), (3.87, 43800.0), (7.53, 22250.0), (14.35, 11800.0), (28.51, 5940.0), (57.01, 2970.0), (114.41, 1480), (206.34, 820.0)]
+def calc_pll_pump(fref, fvco):
+    fclkin_idx = int((fref - 1) // 30)
+    if (fclkin_idx == 13 and fref <= 395) or (fclkin_idx == 14 and fref <= 430) or (fclkin_idx == 15 and fref <= 465) or fclkin_idx == 16:
+        fclkin_idx = fclkin_idx - 1
+
+    r_vals = [(fr[1], len(_freq_R) - 1 - idx) for idx, fr in enumerate(_freq_R) if fr[0] < fref]
+    r_vals.reverse()
+
+    # Find the resistor that provides the minimum current through the capacitor
+    K0 = (497.5 - math.sqrt(247506.25 - (2675.4 - fvco) * 78.46)) / 39.23
+    K1 = 4.8714 * K0 * K0 + 6.5257 * K0 + 142.67
+    Kvco = 1000000.0 * K1
+    Ndiv = fvco / fref
+    C1 = 6.69244e-11
+
+    for R1, r_idx in r_vals:
+        Ic = (1.8769 / (R1 * R1 * Kvco * C1)) * 4.0 * Ndiv
+        if Ic <= 0.00028:
+            icp = int(Ic * 100000.0 + 0.5) * 10
+            break
+
+    return ((fclkin_idx + 1) * 16, icp, r_idx)
+
+# add the default pll attributes according to the documentation
+_default_pll_inattrs = {
+            'FCLKIN'        : '100.00',
+            'IDIV_SEL'      : '0',
+            'DYN_IDIV_SEL'  : 'false',
+            'FBDIV_SEL'     : '00000000000000000000000000000000',
+            'DYN_FBDIV_SEL' : 'false',
+            'ODIV_SEL'      : '00000000000000000000000000001000',
+            'DYN_ODIV_SEL'  : 'false',
+            'PSDA_SEL'      : '0000 ', # XXX extra space for compatibility, but it will work with or without it in the future
+            'DUTYDA_SEL'    : '1000 ', # ^^^
+            'DYN_DA_EN'     : 'false',
+            'CLKOUT_FT_DIR' : '1',
+            'CLKOUT_DLY_STEP': '00000000000000000000000000000000',
+            'CLKOUTP_FT_DIR': '1',
+            'CLKOUTP_DLY_STEP': '00000000000000000000000000000000',
+            'DYN_SDIV_SEL'  : '00000000000000000000000000000010',
+            'CLKFB_SEL'     : 'internal',
+            'CLKOUTD_SRC'   : 'CLKOUT',
+            'CLKOUTD3_SRC'  : 'CLKOUT',
+            'CLKOUT_BYPASS' : 'false',
+            'CLKOUTP_BYPASS': 'false',
+            'CLKOUTD_BYPASS': 'false',
+            'DEVICE'        : 'GW1N-1'
+
+        }
+
+def add_pll_default_attrs(attrs):
+    pll_inattrs = attrs.copy()
+    for k, v in _default_pll_inattrs.items():
+        if k in pll_inattrs.keys():
+            continue
+        pll_inattrs[k] = v
+    return pll_inattrs
+
+_default_pll_internal_attrs = {
+            'INSEL': 'CLKIN1',
+            'FBSEL': 'CLKFB3',
+            'PLOCK': 'ENABLE',
+            'FLOCK': 'ENABLE',
+            'FLTOP': 'ENABLE',
+            'GMCMODE': 15,
+            'CLKOUTDIV3': 'ENABLE',
+            'CLKOUTDIV': 'ENABLE',
+            'CLKOUTPS': 'ENABLE',
+            'PDN': 'ENABLE',
+            'PASEL': 0,
+            'IRSTEN': 'DISABLE',
+            'SRSTEN': 'DISABLE',
+            'PWDEN': 'ENABLE',
+            'RSTEN': 'ENABLE',
+            'FLDCOUNT': 16,
+            'GMCGAIN': 0,
+            'LPR': 'R4',
+            'ICPSEL': 50,
+}
+
+# typ - PLL type (RPLL, etc)
+def set_pll_attrs(db, typ, attrs):
+    pll_inattrs = add_pll_default_attrs(attrs)
+    pll_attrs = _default_pll_internal_attrs.copy()
+
+    if typ not in ['RPLL']:
+        raise Exception(f"PLL type {typ} is not supported for now")
+
+    # parse attrs
+    for attr, val in pll_inattrs.items():
+        if attr in pll_attrs.keys():
+            pll_attrs[attr] = val
+        if attr == 'CLKOUTD_SRC':
+            if val == 'CLKOUTP':
+                pll_attrs['CLKOUTDIVSEL'] = 'CLKOUTPS'
+            continue
+        if attr == 'CLKOUTD3_SRC':
+            if val == 'CLKOUTP':
+                pll_attrs['CLKOUTDIV3SEL'] = 'CLKOUTPS'
+            continue
+        if attr == 'DYN_IDIV_SEL':
+            if val == 'true':
+                pll_attrs['IDIVSEL'] = 'DYN'
+            continue
+        if attr == 'DYN_FBDIV_SEL':
+            if val == 'true':
+                pll_attrs['FDIVSEL'] = 'DYN'
+            continue
+        if attr == 'DYN_ODIV_SEL':
+            if val == 'true':
+                pll_attrs['ODIVSEL'] = 'DYN'
+            continue
+        if attr == 'CLKOUT_BYPASS':
+            if val == 'true':
+                pll_attrs['BYPCK'] = 'BYPASS'
+            continue
+        if attr == 'CLKOUTP_BYPASS':
+            if val == 'true':
+                pll_attrs['BYPCKPS'] = 'BYPASS'
+            continue
+        if attr == 'CLKOUTD_BYPASS':
+            if val == 'true':
+                pll_attrs['BYPCKDIV'] = 'BYPASS'
+            continue
+        if attr == 'IDIV_SEL':
+            idiv = 1 + int(val, 2)
+            pll_attrs['IDIV'] = idiv
+            continue
+        if attr == 'FBDIV_SEL':
+            fbdiv = 1 + int(val, 2)
+            pll_attrs['FDIV'] = fbdiv
+            continue
+        if attr == 'DYN_SDIV_SEL':
+            pll_attrs['SDIV'] = int(val, 2)
+            continue
+        if attr == 'ODIV_SEL':
+            odiv = int(val, 2)
+            pll_attrs['ODIV'] = odiv
+            continue
+        if attr == 'DYN_DA_EN':
+            if val == 'true':
+                pll_attrs['DPSEL'] = 'DYN'
+                pll_attrs['DUTY'] = 0
+                pll_attrs['PHASE'] = 0
+                pll_attrs['PASEL'] = 'DISABLE'
+                # steps in 50ps
+                tmp_val = int(pll_inattrs['CLKOUT_DLY_STEP'], 2) * 50
+                pll_attrs['OPDLY'] = tmp_val
+                # XXX here is unclear according to the documentation only three
+                # values are allowed: 0, 1 and 2, but there are 4 fuses (0, 50,
+                # 75, 100). Find out what to do with 75
+                tmp_val = int(pll_inattrs['CLKOUTP_DLY_STEP'], 2) * 50
+                pll_attrs['OSDLY'] = tmp_val
+            else:
+                pll_attrs['OSDLY'] = 'DISABLE'
+                pll_attrs['OPDLY'] = 'DISABLE'
+                phase_val = int(pll_inattrs['PSDA_SEL'].strip(), 2)
+                pll_attrs['PHASE'] = phase_val
+                duty_val = int(pll_inattrs['DUTYDA_SEL'].strip(), 2)
+                # XXX there are fuses for 15 variants (excluding 0) so for now
+                # we will implement all of them, including those prohibited by
+                # documentation 1 and 15
+                if (phase_val + duty_val) < 16:
+                    duty_val = phase_val + duty_val
+                else:
+                    duty_val = phase_val + duty_val - 16
+                pll_attrs['DUTY'] = duty_val
+            continue
+        if attr == 'FCLKIN':
+            fclkin = float(val)
+            if fclkin < 3 or fclkin > _permitted_freqs[device][0]:
+                print(f"The {fclkin}MHz frequency is outside the permissible range of 3-{_permitted_freqs[device][0]}MHz.")
+                fclkin = 100.0
+            continue
+
+    # static vs dynamic
+    if pll_inattrs['DYN_IDIV_SEL'] == 'false' and pll_inattrs['DYN_FBDIV_SEL'] == 'false' and pll_inattrs['DYN_ODIV_SEL'] == 'false':
+        # static. We can immediately check the compatibility of the divisors
+        clkout = fclkin * fbdiv / idiv
+        if clkout <= _permitted_freqs[device][2] or clkout > _permitted_freqs[device][1]:
+            raise Exception(f"CLKOUT = FCLKIN*(FBDIV_SEL+1)/(IDIV_SEL+1) = {clkout}MHz not in range {_permitted_freqs[device][2]} - {_permitted_freqs[device][1]}MHz")
+        pfd = fclkin / idiv
+        if pfd < 3.0 or pfd > _permitted_freqs[device][0]:
+            raise Exception(f"PFD = FCLKIN/(IDIV_SEL+1) = {pfd}MHz not in range 3.0 - {_permitted_freqs[device][0]}MHz")
+        fvco = odiv * fclkin * fbdiv / idiv
+        if fvco < _permitted_freqs[device][4] or  fvco > _permitted_freqs[device][3]:
+            raise Exception(f"VCO = FCLKIN*(FBDIV_SEL+1)*ODIV_SEL/(IDIV_SEL+1) = {fvco}MHz not in range {_permitted_freqs[device][4]} - {_permitted_freqs[device][3]}MHz")
+
+    # pump
+    fref = fclkin / idiv
+    fvco = (odiv * fbdiv * fclkin) / idiv
+    fclkin_idx, icp, r_idx = calc_pll_pump(fref, fvco)
+
+    pll_attrs['FLDCOUNT'] = fclkin_idx
+    pll_attrs['ICPSEL'] = int(icp)
+    pll_attrs['LPR'] = f"R{r_idx}"
+
+    fin_attrs = set()
+    for attr, val in pll_attrs.items():
+        if isinstance(val, str):
+            val = pll_attrvals[val]
+        add_attr_val(db, 'PLL', fin_attrs, pll_attrids[attr], val)
+    #print(fin_attrs)
+    return fin_attrs
+
 iostd_alias = {
         "HSTL18_II"  : "HSTL18_I",
         "SSTL18_I"   : "HSTL18_I",
@@ -91,7 +329,9 @@ def place(db, tilemap, bels, cst, args):
         tile = tilemap[(row-1, col-1)]
         if typ == "GSR":
             pass
-        if typ == "BUFS":
+        elif typ.startswith('MUX2_'):
+            pass
+        elif typ == "BUFS":
             # fuses must be reset in order to activate so remove them
             bits2zero = set()
             for fuses in [fuses for fuses in parms.keys() if fuses in {'L', 'R'}]:
@@ -99,7 +339,7 @@ def place(db, tilemap, bels, cst, args):
             for r, c in bits2zero:
                 tile[r][c] = 0
 
-        if typ in {'OSC', 'OSCZ', 'OSCF', 'OSCH'}:
+        elif typ in {'OSC', 'OSCZ', 'OSCF', 'OSCH'}:
             divisor = int(parms['FREQ_DIV'], 2)
             if divisor % 2 == 1:
                 raise Exception(f"Divisor of {typ} must be even")
@@ -108,7 +348,7 @@ def place(db, tilemap, bels, cst, args):
                 bits = tiledata.bels[typ].modes[divisor]
                 for r, c in bits:
                     tile[r][c] = 1
-        if typ == "SLICE":
+        elif typ == "SLICE":
             lutmap = tiledata.bels[f'LUT{num}'].flags
 
             if 'ALU_MODE' in parms.keys():
@@ -257,6 +497,21 @@ def place(db, tilemap, bels, cst, args):
                 bits.update(bel.flags['IOBUF'])
             for r, c in bits:
                 tile[r][c] = 1
+        elif typ == "RAMW":
+            bel = tiledata.bels['RAM16']
+            bits = bel.modes['0']
+            #print(typ, bits)
+            for r, c in bits:
+                tile[r][c] = 1
+        elif typ.startswith('RPLL'):
+            pll_attrs = set_pll_attrs(db, 'RPLL', parms)
+            bits = get_shortval_fuses(db, tiledata.ttyp, pll_attrs, 'PLL')
+            #print(typ, bits)
+            for r, c in bits:
+                tile[r][c] = 1
+
+        else:
+            print("unknown type", typ)
 
     # If the entire bank has only inputs, the LVCMOS12/15/18 bit is set
     # in each IBUF regardless of the actual I/O standard.
@@ -347,6 +602,12 @@ def dualmode_pins(db, tilemap, args):
             tile[row][col] = 1
 
 def main():
+    global device
+    pil_available = True
+    try:
+        from PIL import Image
+    except ImportError:
+        pil_available = False
     parser = argparse.ArgumentParser(description='Pack Gowin bitstream')
     parser.add_argument('netlist')
     parser.add_argument('-d', '--device', required=True)
@@ -360,7 +621,8 @@ def main():
     parser.add_argument('--ready_as_gpio', action = 'store_true')
     parser.add_argument('--done_as_gpio', action = 'store_true')
     parser.add_argument('--reconfign_as_gpio', action = 'store_true')
-    parser.add_argument('--png')
+    if pil_available:
+        parser.add_argument('--png')
 
     args = parser.parse_args()
     device = args.device
@@ -385,7 +647,7 @@ def main():
     dualmode_pins(db, tilemap, args)
     res = chipdb.fuse_bitmap(db, tilemap)
     header_footer(db, res, args.compress)
-    if args.png:
+    if pil_available and args.png:
         bslib.display(args.png, res)
     bslib.write_bitstream(args.output, res, db.cmd_hdr, db.cmd_ftr, args.compress)
     if args.cst:
